/**
 * Orthanc - A Lightweight, RESTful DICOM Store
 * Copyright (C) 2012-2016 Sebastien Jodogne, Medical Physics
 * Department, University Hospital of Liege, Belgium
 * Copyright (C) 2017-2019 Osimis S.A., Belgium
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License
 * as published by the Free Software Foundation, either version 3 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 **/


#pragma once

#include "InMemoryTiledImage.h"

namespace OrthancWSI
{
  class TruncatedPyramidWriter : public IPyramidWriter
  {
  private:
    IPyramidWriter&     lowerLevels_;
    InMemoryTiledImage  upperLevel_;
    unsigned int        upperLevelIndex_;
    
  public:
    TruncatedPyramidWriter(IPyramidWriter& lower,
                           unsigned int upperLevelIndex);

    virtual unsigned int GetLevelCount() const
    {
      return upperLevelIndex_ + 1;
    }

    virtual Orthanc::PixelFormat GetPixelFormat() const 
    {
      return lowerLevels_.GetPixelFormat();
    }

    virtual unsigned int GetTileWidth() const
    {
      return lowerLevels_.GetTileWidth();
    }

    virtual unsigned int GetTileHeight() const 
    {
      return lowerLevels_.GetTileHeight();
    }

    virtual unsigned int GetCountTilesX(unsigned int level) const;

    virtual unsigned int GetCountTilesY(unsigned int level) const;

    virtual void WriteRawTile(const std::string& tile,
                              ImageCompression compression,
                              unsigned int level,
                              unsigned int x,
                              unsigned int y);

    virtual void EncodeTile(const Orthanc::ImageAccessor& tile,
                            unsigned int level,
                            unsigned int x, 
                            unsigned int y);

    InMemoryTiledImage& GetUpperLevel()
    {
      return upperLevel_;
    }
  };
}
